#include <hidef.h>      /* common defines and macros */
#include <mc9s12dp256.h>     /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12dp256b"

/*
 * Protocol
 */
enum Command { 
  // Lights
    LOnL, LOffL, // Turn on/off Line Change Left
    LOnR, LOffR, // Turn on/off Line Change Right
    LOnW, LOffW, // Turn on/off Warning Lights
  // State
    SAOn, SAOff, // Automatically send the state on/off
    SG,          // Give the state
  // Acceleration
    ASV,         // Set the acceleration value
    AIV,         // Increase acceleration value
    ADV          // Decrease acceleration value
};

/*
 * PACKET 
 * 0 ID: (1 byte)
 *   see below
 * 1 LENGTH: (1 byte)
 *   length of data part
 * 2-6 DATA: (0 - 5 bytes)
 * 7 CHECKSUM: (1 byte)
 *   sum of all bytes (as unsigned char with possible overrun)
 */
unsigned char packet[2][8];
unsigned char *packet_buffer = packet[0];
unsigned char *packet_current = packet[1];
unsigned char packet_buffer_idx = 0; //current receive index
unsigned char is_packet_ready = 0; //packet is ready

#pragma CODE_SEG __NEAR_SEG NON_BANKED
/*serial interrupt: collect characters and fill "packet" array accordingly*/
interrupt void SCI0_ISR(void){
  unsigned char *packet_tmp; // used to exchange current and buffer pointers

  if(SCI0SR1 & 0x20) { // ready to receive data: RDRF is set
    if(packet_buffer_idx > 1/*not ID or LENGTH*/
    && packet_buffer_idx - 1 > packet_buffer[1]/*is checksum*/) {
      packet_buffer_idx = 7; // put checksum always on idx 7
    }
        
    packet_buffer[packet_buffer_idx] = SCI0DRL;
    
    if(packet_buffer_idx == 7) { // is checksum
      packet_tmp = packet_current; // exchange pointers
      packet_current = packet_buffer;
      packet_buffer = packet_tmp;
      is_packet_ready = 1; // mark packet as finished
      packet_buffer_idx = 0; // start new packet
      
    } else { // prepare for next character
      packet_buffer_idx++;
    }
  }
}

/*initialize the packet buffer*/
void init_packet(unsigned char *packet) {
  unsigned char i = 0;
  for(; i<8; i++) {
    packet[i] = 0;    
  }  
}

/*check checksum*/
unsigned char is_checksum_ok(unsigned char *packet) {
  unsigned char sum = 0;
  unsigned char i = 0;
  for(; i<7; i++) {
    sum += packet[i];
  }
  return (sum == packet[7]/*checksum*/);
}

/*
 * TIMER
 */

/* finite state machine for Line Change/Warning Lights */
enum Light_State { 
  /*-1*/inv=-1,
  /* 0*/NO,
  /* 1*/LEFT,
  /* 2*/RIGHT,
  /* 3*/WARN,
  /* 4*/LEFT_WARN,
  /* 5*/RIGHT_WARN
};
enum Light_State light_state_command_map[6][6] = {
                   /*LOnL       LOffL LOnR        LOffR LOnW        LOffW*/
  /*0 NO*/         { LEFT,      inv,  RIGHT,      inv,  WARN,       inv   },
  /*1 LEFT*/       { inv,       NO,   inv,        inv,  LEFT_WARN,  inv   },
  /*2 RIGHT*/      { inv,       inv,  inv,        NO,   RIGHT_WARN, inv   },
  /*3 WARN*/       { LEFT_WARN, inv,  RIGHT_WARN, inv,  inv,        NO    },
  /*4 LEFT_WARN*/  { inv,       WARN, inv,        inv,  inv,        LEFT  },
  /*5 RIGHT_WARN*/ { inv,       inv,  inv,        WARN, inv,        RIGHT }
};
enum Light_State light_state = NO; // no blinking
enum Light_State last_light_state = NO; // no blinking
void light_fsn() {
  last_light_state = light_state;
  switch(light_state) {
  case NO: // no blinking
    PORTB |= 0x81;
    break;
  case LEFT: // left blinking (P2.15)
    PORTB ^= 0x80;
    PORTB |= 0x01;
    break;
  case RIGHT: // right blinking (P2.8)
    PORTB ^= 0x01;
    PORTB |= 0x80;
    break;
  case WARN: // warning lights (both)
  case LEFT_WARN: // left blinking + warning
  case RIGHT_WARN: // right blinking + warning
    PORTB ^= 0x81;
    break;
  }
}

/*
 * STATE
 */
void send_state() {
  
      //TODO: send packet
      //SCI0DRL = (unsigned char)p_command;
      //while(!(SCI0SR1 & 0x80)); //wait till output buffer is empty
}
/* finite state machine for sending state automatically */
unsigned char send_state_auto = 0; // do not send
void state_fsn() {
  switch(send_state_auto) {
    case 1:
      send_state();
      break;
    case 0:
      break;
  }
}

#pragma CODE_SEG __NEAR_SEG NON_BANKED

/*timer interrupt*/
unsigned int timer_cnt = 0;
interrupt void RTI_ISR(void){
  CRGFLG = 0x80; //ack interrupt
  ++timer_cnt;
  
  if(last_light_state != light_state && last_light_state == NO) {
    // turn lights on: always start with short cyle
    timer_cnt = 16;
  }
  //duty cycle: 1 Hz; 70 % duty cycle
  if(timer_cnt == 5 || timer_cnt == 16) {
    //5*65ms = 325ms
    //11*65ms = 715ms
    light_fsn();
  }
  if(timer_cnt == 16 /*=5+11*/) {
    timer_cnt=0; // new cycle
  }  
}


/*
 * ACCELERATION
 * set the speed to a value between 0 and 239
 */
unsigned char accelerationValue = 0;
void setAcceleration(unsigned char value) {
  if(value > 239) value = 239;
  accelerationValue = value;
  PORTB |= 0x7E;
  PORTB &= 0x81 | (255 << ((value + 1) / 40 + 2));
}

/*
 * INIT
 */
#pragma CODE_SEG DEFAULT
void Init(){
 //for serial
 SCI0BDL = (unsigned char)((16000000/2)/19200/16);
 SCI0CR2 = 0x2C;
 init_packet(packet_buffer);
 init_packet(packet_current);
 
 //for timer
 //RTICTL = 0x1F;//Real Time Interrupt Control Register = 1.024 ms
 RTICTL = 0x7F; //Real Time Interrupt Control Register = 65.536 ms
 CRGINT = 0x80;//CRG Interrupt Enable Register
 
 EnableInterrupts;
 
 //for LEDs
 DDRB = 0xFF;//Data Direction Register set
 DDRE = 0x0;
 PORTB = 0xFF; //Set LEDs off
 
 //acceleration
 setAcceleration(0);
}


/*
 * MAIN
 */
void main(void) {
  Init();
    
  for(;;) { 
    if(is_packet_ready && is_checksum_ok(packet_current)) {      
      /* send ack */
      enum Command p_command = packet_current[0];
      
      //TODO: generate full packet as ack
      SCI0DRL = (unsigned char)p_command;
      while(!(SCI0SR1 & 0x80)); //wait till output buffer is empty
      
      switch(p_command) {
      /* Lights */
      case LOnL: // Turn on Line Change Left
      case LOffL: // Turn off Line Change Left
      case LOnR: // Turn on Line Change Right
      case LOffR: // Turn off Line Change Right
      case LOnW: // Turn on Warning Lights
      case LOffW: // Turn off Warning Lights
        if(light_state_command_map[light_state][p_command] != inv) {
          light_state = light_state_command_map[light_state][p_command];         
        }
        break;
      
      /* State */
      case SAOn: // Automatically send the state on
        send_state_auto = 1;
        break;
      case SAOff: // Automatically send the state off
        send_state_auto = 0;
        break;
      case SG: // Give the state
        send_state();
        break;
      /* Acceleration */
      case ASV: // Set the acceleration value
        setAcceleration(packet_current[2]);
        break;
      case AIV: // Increase acceleration value
        if(239 - accelerationValue < packet_current[2]) { // overflow detection
          setAcceleration(239);
        } else {  
          setAcceleration(accelerationValue + packet_current[2]);
        }
        break;
      case ADV: // Decrease acceleration value
        if(accelerationValue < packet_current[2]) { // underflow detection
          setAcceleration(0);
        } else {
          setAcceleration(accelerationValue - packet_current[2]);
        }
        break;
      }
      init_packet(packet_current);
      is_packet_ready = 0;      
    }
  } /* wait forever */
  /* please make sure that you never leave this function */
}